@@ -5,6 +5,10 @@
 package js_parser
 
 import (
+	"fmt"
+	"strings"
+
+	"github.com/evanw/esbuild/internal/compat"
 	"github.com/evanw/esbuild/internal/js_ast"
 	"github.com/evanw/esbuild/internal/js_lexer"
 	"github.com/evanw/esbuild/internal/logger"
@@ -156,7 +160,8 @@ func (p *parser) skipTypeScriptType(level js_ast.L) {
 }
 
 type skipTypeOpts struct {
-	isReturnType bool
+	isReturnType     bool
+	allowTupleLabels bool
 }
 
 type tsTypeIdentifierKind uint8
@@ -197,9 +202,18 @@ func (p *parser) skipTypeScriptTypeWithOpts(level js_ast.L, opts skipTypeOpts) {
 		switch p.lexer.Token {
 		case js_lexer.TNumericLiteral, js_lexer.TBigIntegerLiteral, js_lexer.TStringLiteral,
 			js_lexer.TNoSubstitutionTemplateLiteral, js_lexer.TTrue, js_lexer.TFalse,
-			js_lexer.TNull, js_lexer.TVoid, js_lexer.TConst:
+			js_lexer.TNull, js_lexer.TVoid:
 			p.lexer.Next()
 
+		case js_lexer.TConst:
+			r := p.lexer.Range()
+			p.lexer.Next()
+
+			// "[const: number]"
+			if opts.allowTupleLabels && p.lexer.Token == js_lexer.TColon {
+				p.log.Add(logger.Error, &p.tracker, r, "Unexpected \"const\"")
+			}
+
 		case js_lexer.TThis:
 			p.lexer.Next()
 
@@ -229,6 +243,12 @@ func (p *parser) skipTypeScriptTypeWithOpts(level js_ast.L, opts skipTypeOpts) {
 		case js_lexer.TImport:
 			// "import('fs')"
 			p.lexer.Next()
+
+			// "[import: number]"
+			if opts.allowTupleLabels && p.lexer.Token == js_lexer.TColon {
+				return
+			}
+
 			p.lexer.Expect(js_lexer.TOpenParen)
 			p.lexer.Expect(js_lexer.TStringLiteral)
 			p.lexer.Expect(js_lexer.TCloseParen)
@@ -237,6 +257,12 @@ func (p *parser) skipTypeScriptTypeWithOpts(level js_ast.L, opts skipTypeOpts) {
 			// "new () => Foo"
 			// "new <T>() => Foo<T>"
 			p.lexer.Next()
+
+			// "[new: number]"
+			if opts.allowTupleLabels && p.lexer.Token == js_lexer.TColon {
+				return
+			}
+
 			p.skipTypeScriptTypeParameters()
 			p.skipTypeScriptParenOrFnType()
 
@@ -304,6 +330,12 @@ func (p *parser) skipTypeScriptTypeWithOpts(level js_ast.L, opts skipTypeOpts) {
 
 		case js_lexer.TTypeof:
 			p.lexer.Next()
+
+			// "[typeof: number]"
+			if opts.allowTupleLabels && p.lexer.Token == js_lexer.TColon {
+				return
+			}
+
 			if p.lexer.Token == js_lexer.TImport {
 				// "typeof import('fs')"
 				continue
@@ -330,7 +362,7 @@ func (p *parser) skipTypeScriptTypeWithOpts(level js_ast.L, opts skipTypeOpts) {
 				if p.lexer.Token == js_lexer.TDotDotDot {
 					p.lexer.Next()
 				}
-				p.skipTypeScriptType(js_ast.LLowest)
+				p.skipTypeScriptTypeWithOpts(js_ast.LLowest, skipTypeOpts{allowTupleLabels: true})
 				if p.lexer.Token == js_lexer.TQuestion {
 					p.lexer.Next()
 				}
@@ -361,6 +393,18 @@ func (p *parser) skipTypeScriptTypeWithOpts(level js_ast.L, opts skipTypeOpts) {
 			}
 
 		default:
+			// "[function: number]"
+			if opts.allowTupleLabels && p.lexer.IsIdentifierOrKeyword() {
+				if p.lexer.Token != js_lexer.TFunction {
+					p.log.Add(logger.Error, &p.tracker, p.lexer.Range(), fmt.Sprintf("Unexpected %q", p.lexer.Raw()))
+				}
+				p.lexer.Next()
+				if p.lexer.Token != js_lexer.TColon {
+					p.lexer.Expect(js_lexer.TColon)
+				}
+				return
+			}
+
 			p.lexer.Unexpected()
 		}
 		break
@@ -694,6 +738,28 @@ func (p *parser) trySkipTypeScriptArrowArgsWithBacktracking() bool {
 	return true
 }
 
+// Returns true if the current less-than token is considered to be an arrow
+// function under TypeScript's rules for files containing JSX syntax
+func (p *parser) isTSArrowFnJSX() (isTSArrowFn bool) {
+	oldLexer := p.lexer
+	p.lexer.Next()
+
+	// Look ahead to see if this should be an arrow function instead
+	if p.lexer.Token == js_lexer.TIdentifier {
+		p.lexer.Next()
+		if p.lexer.Token == js_lexer.TComma {
+			isTSArrowFn = true
+		} else if p.lexer.Token == js_lexer.TExtends {
+			p.lexer.Next()
+			isTSArrowFn = p.lexer.Token != js_lexer.TEquals && p.lexer.Token != js_lexer.TGreaterThan
+		}
+	}
+
+	// Restore the lexer
+	p.lexer = oldLexer
+	return
+}
+
 // This function is taken from the official TypeScript compiler source code:
 // https://github.com/microsoft/TypeScript/blob/master/src/compiler/parser.ts
 func (p *parser) canFollowTypeArgumentsInExpression() bool {
@@ -831,26 +897,51 @@ func (p *parser) parseTypeScriptEnumStmt(loc logger.Loc, opts parseStmtOpts) js_
 	nameText := p.lexer.Identifier
 	p.lexer.Expect(js_lexer.TIdentifier)
 	name := js_ast.LocRef{Loc: nameLoc, Ref: js_ast.InvalidRef}
-	argRef := js_ast.InvalidRef
+
+	// Generate the namespace object
+	exportedMembers := p.getOrCreateExportedNamespaceMembers(nameText, opts.isExport)
+	tsNamespace := &js_ast.TSNamespaceScope{
+		ExportedMembers: exportedMembers,
+		ArgRef:          js_ast.InvalidRef,
+		IsEnumScope:     true,
+	}
+	enumMemberData := &js_ast.TSNamespaceMemberNamespace{
+		ExportedMembers: exportedMembers,
+	}
+
+	// Declare the enum and create the scope
 	if !opts.isTypeScriptDeclare {
 		name.Ref = p.declareSymbol(js_ast.SymbolTSEnum, nameLoc, nameText)
 		p.pushScopeForParsePass(js_ast.ScopeEntry, loc)
+		p.currentScope.TSNamespace = tsNamespace
+		p.refToTSNamespaceMemberData[name.Ref] = enumMemberData
 	}
-	p.lexer.Expect(js_lexer.TOpenBrace)
 
+	p.lexer.Expect(js_lexer.TOpenBrace)
 	values := []js_ast.EnumValue{}
 
+	oldFnOrArrowData := p.fnOrArrowDataParse
+	p.fnOrArrowDataParse = fnOrArrowDataParse{
+		isThisDisallowed: true,
+		needsAsyncLoc:    logger.Loc{Start: -1},
+	}
+
+	// Parse the body
 	for p.lexer.Token != js_lexer.TCloseBrace {
+		nameRange := p.lexer.Range()
 		value := js_ast.EnumValue{
-			Loc: p.lexer.Loc(),
+			Loc: nameRange.Loc,
 			Ref: js_ast.InvalidRef,
 		}
 
 		// Parse the name
+		var nameText string
 		if p.lexer.Token == js_lexer.TStringLiteral {
 			value.Name = p.lexer.StringLiteral()
+			nameText = js_lexer.UTF16ToString(value.Name)
 		} else if p.lexer.IsIdentifierOrKeyword() {
-			value.Name = js_lexer.StringToUTF16(p.lexer.Identifier)
+			nameText = p.lexer.Identifier
+			value.Name = js_lexer.StringToUTF16(nameText)
 		} else {
 			p.lexer.Expect(js_lexer.TIdentifier)
 		}
@@ -869,12 +960,44 @@ func (p *parser) parseTypeScriptEnumStmt(loc logger.Loc, opts parseStmtOpts) js_
 
 		values = append(values, value)
 
+		// Add this enum value as a member of the enum's namespace
+		exportedMembers[nameText] = js_ast.TSNamespaceMember{
+			Loc:         value.Loc,
+			Data:        &js_ast.TSNamespaceMemberProperty{},
+			IsEnumValue: true,
+		}
+
 		if p.lexer.Token != js_lexer.TComma && p.lexer.Token != js_lexer.TSemicolon {
+			if p.lexer.IsIdentifierOrKeyword() || p.lexer.Token == js_lexer.TStringLiteral {
+				var errorLoc logger.Loc
+				var errorText string
+
+				if value.ValueOrNil.Data == nil {
+					errorLoc = logger.Loc{Start: nameRange.End()}
+					errorText = fmt.Sprintf("Expected \",\" after %q in enum", nameText)
+				} else {
+					var nextName string
+					if p.lexer.Token == js_lexer.TStringLiteral {
+						nextName = js_lexer.UTF16ToString(p.lexer.StringLiteral())
+					} else {
+						nextName = p.lexer.Identifier
+					}
+					errorLoc = p.lexer.Loc()
+					errorText = fmt.Sprintf("Expected \",\" before %q in enum", nextName)
+				}
+
+				data := p.tracker.MsgData(logger.Range{Loc: errorLoc}, errorText)
+				data.Location.Suggestion = ","
+				p.log.AddMsg(logger.Msg{Kind: logger.Error, Data: data})
+				panic(js_lexer.LexerPanic{})
+			}
 			break
 		}
 		p.lexer.Next()
 	}
 
+	p.fnOrArrowDataParse = oldFnOrArrowData
+
 	if !opts.isTypeScriptDeclare {
 		// Avoid a collision with the enum closure argument variable if the
 		// enum exports a symbol with the same name as the enum itself:
@@ -909,11 +1032,12 @@ func (p *parser) parseTypeScriptEnumStmt(loc logger.Loc, opts parseStmtOpts) js_
 			// Add a "_" to make tests easier to read, since non-bundler tests don't
 			// run the renamer. For external-facing things the renamer will avoid
 			// collisions automatically so this isn't important for correctness.
-			argRef = p.newSymbol(js_ast.SymbolHoisted, "_"+nameText)
-			p.currentScope.Generated = append(p.currentScope.Generated, argRef)
+			tsNamespace.ArgRef = p.newSymbol(js_ast.SymbolHoisted, "_"+nameText)
+			p.currentScope.Generated = append(p.currentScope.Generated, tsNamespace.ArgRef)
 		} else {
-			argRef = p.declareSymbol(js_ast.SymbolHoisted, nameLoc, nameText)
+			tsNamespace.ArgRef = p.declareSymbol(js_ast.SymbolHoisted, nameLoc, nameText)
 		}
+		p.refToTSNamespaceMemberData[tsNamespace.ArgRef] = enumMemberData
 
 		p.popScope()
 	}
@@ -930,7 +1054,7 @@ func (p *parser) parseTypeScriptEnumStmt(loc logger.Loc, opts parseStmtOpts) js_
 
 	return js_ast.Stmt{Loc: loc, Data: &js_ast.SEnum{
 		Name:     name,
-		Arg:      argRef,
+		Arg:      tsNamespace.ArgRef,
 		Values:   values,
 		IsExport: opts.isExport,
 	}}
@@ -991,17 +1115,65 @@ func (p *parser) parseTypeScriptImportEqualsStmt(loc logger.Loc, opts parseStmtO
 	}}
 }
 
+// Generate a TypeScript namespace object for this namespace's scope. If this
+// namespace is another block that is to be merged with an existing namespace,
+// use that earlier namespace's object instead.
+func (p *parser) getOrCreateExportedNamespaceMembers(name string, isExport bool) js_ast.TSNamespaceMembers {
+	// Merge with a sibling namespace from the same scope
+	if existingMember, ok := p.currentScope.Members[name]; ok {
+		if memberData, ok := p.refToTSNamespaceMemberData[existingMember.Ref]; ok {
+			if nsMemberData, ok := memberData.(*js_ast.TSNamespaceMemberNamespace); ok {
+				return nsMemberData.ExportedMembers
+			}
+		}
+	}
+
+	// Merge with a sibling namespace from a different scope
+	if isExport {
+		if parentNamespace := p.currentScope.TSNamespace; parentNamespace != nil {
+			if existing, ok := parentNamespace.ExportedMembers[name]; ok {
+				if existing, ok := existing.Data.(*js_ast.TSNamespaceMemberNamespace); ok {
+					return existing.ExportedMembers
+				}
+			}
+		}
+	}
+
+	// Otherwise, generate a new namespace object
+	return make(js_ast.TSNamespaceMembers)
+}
+
 func (p *parser) parseTypeScriptNamespaceStmt(loc logger.Loc, opts parseStmtOpts) js_ast.Stmt {
 	// "namespace Foo {}"
 	nameLoc := p.lexer.Loc()
 	nameText := p.lexer.Identifier
 	p.lexer.Next()
 
+	// Generate the namespace object
+	exportedMembers := p.getOrCreateExportedNamespaceMembers(nameText, opts.isExport)
+	tsNamespace := &js_ast.TSNamespaceScope{
+		ExportedMembers: exportedMembers,
+		ArgRef:          js_ast.InvalidRef,
+	}
+	nsMemberData := &js_ast.TSNamespaceMemberNamespace{
+		ExportedMembers: exportedMembers,
+	}
+
+	// Declare the namespace and create the scope
 	name := js_ast.LocRef{Loc: nameLoc, Ref: js_ast.InvalidRef}
 	scopeIndex := p.pushScopeForParsePass(js_ast.ScopeEntry, loc)
+	p.currentScope.TSNamespace = tsNamespace
 
 	oldHasNonLocalExportDeclareInsideNamespace := p.hasNonLocalExportDeclareInsideNamespace
+	oldFnOrArrowData := p.fnOrArrowDataParse
 	p.hasNonLocalExportDeclareInsideNamespace = false
+	p.fnOrArrowDataParse = fnOrArrowDataParse{
+		isThisDisallowed:   true,
+		isReturnDisallowed: true,
+		needsAsyncLoc:      logger.Loc{Start: -1},
+	}
+
+	// Parse the statements inside the namespace
 	var stmts []js_ast.Stmt
 	if p.lexer.Token == js_lexer.TDot {
 		dotLoc := p.lexer.Loc()
@@ -1021,8 +1193,75 @@ func (p *parser) parseTypeScriptNamespaceStmt(loc logger.Loc, opts parseStmtOpts
 		})
 		p.lexer.Next()
 	}
+
 	hasNonLocalExportDeclareInsideNamespace := p.hasNonLocalExportDeclareInsideNamespace
 	p.hasNonLocalExportDeclareInsideNamespace = oldHasNonLocalExportDeclareInsideNamespace
+	p.fnOrArrowDataParse = oldFnOrArrowData
+
+	// Add any exported members from this namespace's body as members of the
+	// associated namespace object.
+	for _, stmt := range stmts {
+		switch s := stmt.Data.(type) {
+		case *js_ast.SFunction:
+			if s.IsExport {
+				name := p.symbols[s.Fn.Name.Ref.InnerIndex].OriginalName
+				member := js_ast.TSNamespaceMember{
+					Loc:  s.Fn.Name.Loc,
+					Data: &js_ast.TSNamespaceMemberProperty{},
+				}
+				exportedMembers[name] = member
+				p.refToTSNamespaceMemberData[s.Fn.Name.Ref] = member.Data
+			}
+
+		case *js_ast.SClass:
+			if s.IsExport {
+				name := p.symbols[s.Class.Name.Ref.InnerIndex].OriginalName
+				member := js_ast.TSNamespaceMember{
+					Loc:  s.Class.Name.Loc,
+					Data: &js_ast.TSNamespaceMemberProperty{},
+				}
+				exportedMembers[name] = member
+				p.refToTSNamespaceMemberData[s.Class.Name.Ref] = member.Data
+			}
+
+		case *js_ast.SNamespace:
+			if s.IsExport {
+				if memberData, ok := p.refToTSNamespaceMemberData[s.Name.Ref]; ok {
+					if nsMemberData, ok := memberData.(*js_ast.TSNamespaceMemberNamespace); ok {
+						member := js_ast.TSNamespaceMember{
+							Loc: s.Name.Loc,
+							Data: &js_ast.TSNamespaceMemberNamespace{
+								ExportedMembers: nsMemberData.ExportedMembers,
+							},
+						}
+						exportedMembers[p.symbols[s.Name.Ref.InnerIndex].OriginalName] = member
+						p.refToTSNamespaceMemberData[s.Name.Ref] = member.Data
+					}
+				}
+			}
+
+		case *js_ast.SEnum:
+			if s.IsExport {
+				if memberData, ok := p.refToTSNamespaceMemberData[s.Name.Ref]; ok {
+					if nsMemberData, ok := memberData.(*js_ast.TSNamespaceMemberNamespace); ok {
+						member := js_ast.TSNamespaceMember{
+							Loc: s.Name.Loc,
+							Data: &js_ast.TSNamespaceMemberNamespace{
+								ExportedMembers: nsMemberData.ExportedMembers,
+							},
+						}
+						exportedMembers[p.symbols[s.Name.Ref.InnerIndex].OriginalName] = member
+						p.refToTSNamespaceMemberData[s.Name.Ref] = member.Data
+					}
+				}
+			}
+
+		case *js_ast.SLocal:
+			if s.IsExport {
+				p.exportDeclsInsideNamespace(exportedMembers, s.Decls)
+			}
+		}
+	}
 
 	// Import assignments may be only used in type expressions, not value
 	// expressions. If this is the case, the TypeScript compiler removes
@@ -1053,7 +1292,6 @@ func (p *parser) parseTypeScriptNamespaceStmt(loc logger.Loc, opts parseStmtOpts
 		return js_ast.Stmt{Loc: loc, Data: &js_ast.STypeScript{}}
 	}
 
-	argRef := js_ast.InvalidRef
 	if !opts.isTypeScriptDeclare {
 		// Avoid a collision with the namespace closure argument variable if the
 		// namespace exports a symbol with the same name as the namespace itself:
@@ -1075,25 +1313,61 @@ func (p *parser) parseTypeScriptNamespaceStmt(loc logger.Loc, opts parseStmtOpts
 			// Add a "_" to make tests easier to read, since non-bundler tests don't
 			// run the renamer. For external-facing things the renamer will avoid
 			// collisions automatically so this isn't important for correctness.
-			argRef = p.newSymbol(js_ast.SymbolHoisted, "_"+nameText)
-			p.currentScope.Generated = append(p.currentScope.Generated, argRef)
+			tsNamespace.ArgRef = p.newSymbol(js_ast.SymbolHoisted, "_"+nameText)
+			p.currentScope.Generated = append(p.currentScope.Generated, tsNamespace.ArgRef)
 		} else {
-			argRef = p.declareSymbol(js_ast.SymbolHoisted, nameLoc, nameText)
+			tsNamespace.ArgRef = p.declareSymbol(js_ast.SymbolHoisted, nameLoc, nameText)
 		}
+		p.refToTSNamespaceMemberData[tsNamespace.ArgRef] = nsMemberData
 	}
 
 	p.popScope()
 	if !opts.isTypeScriptDeclare {
 		name.Ref = p.declareSymbol(js_ast.SymbolTSNamespace, nameLoc, nameText)
+		p.refToTSNamespaceMemberData[name.Ref] = nsMemberData
 	}
 	return js_ast.Stmt{Loc: loc, Data: &js_ast.SNamespace{
 		Name:     name,
-		Arg:      argRef,
+		Arg:      tsNamespace.ArgRef,
 		Stmts:    stmts,
 		IsExport: opts.isExport,
 	}}
 }
 
+func (p *parser) exportDeclsInsideNamespace(exportedMembers js_ast.TSNamespaceMembers, decls []js_ast.Decl) {
+	for _, decl := range decls {
+		p.exportBindingInsideNamespace(exportedMembers, decl.Binding)
+	}
+}
+
+func (p *parser) exportBindingInsideNamespace(exportedMembers js_ast.TSNamespaceMembers, binding js_ast.Binding) {
+	switch b := binding.Data.(type) {
+	case *js_ast.BMissing:
+
+	case *js_ast.BIdentifier:
+		name := p.symbols[b.Ref.InnerIndex].OriginalName
+		member := js_ast.TSNamespaceMember{
+			Loc:  binding.Loc,
+			Data: &js_ast.TSNamespaceMemberProperty{},
+		}
+		exportedMembers[name] = member
+		p.refToTSNamespaceMemberData[b.Ref] = member.Data
+
+	case *js_ast.BArray:
+		for _, item := range b.Items {
+			p.exportBindingInsideNamespace(exportedMembers, item.Binding)
+		}
+
+	case *js_ast.BObject:
+		for _, property := range b.Properties {
+			p.exportBindingInsideNamespace(exportedMembers, property.Value)
+		}
+
+	default:
+		panic("Internal error")
+	}
+}
+
 func (p *parser) generateClosureForTypeScriptNamespaceOrEnum(
 	stmts []js_ast.Stmt, stmtLoc logger.Loc, isExport bool, nameLoc logger.Loc,
 	nameRef js_ast.Ref, argRef js_ast.Ref, stmtsInsideClosure []js_ast.Stmt,
@@ -1108,19 +1382,20 @@ func (p *parser) generateClosureForTypeScriptNamespaceOrEnum(
 	// Make sure to only emit a variable once for a given namespace, since there
 	// can be multiple namespace blocks for the same namespace
 	if (symbol.Kind == js_ast.SymbolTSNamespace || symbol.Kind == js_ast.SymbolTSEnum) && !p.emittedNamespaceVars[nameRef] {
+		decls := []js_ast.Decl{{Binding: js_ast.Binding{Loc: nameLoc, Data: &js_ast.BIdentifier{Ref: nameRef}}}}
 		p.emittedNamespaceVars[nameRef] = true
-		if p.enclosingNamespaceArgRef == nil {
-			// Top-level namespace
+		if p.currentScope == p.moduleScope {
+			// Top-level namespace: "var"
 			stmts = append(stmts, js_ast.Stmt{Loc: stmtLoc, Data: &js_ast.SLocal{
 				Kind:     js_ast.LocalVar,
-				Decls:    []js_ast.Decl{{Binding: js_ast.Binding{Loc: nameLoc, Data: &js_ast.BIdentifier{Ref: nameRef}}}},
+				Decls:    decls,
 				IsExport: isExport,
 			}})
 		} else {
-			// Nested namespace
+			// Nested namespace: "let"
 			stmts = append(stmts, js_ast.Stmt{Loc: stmtLoc, Data: &js_ast.SLocal{
 				Kind:  js_ast.LocalLet,
-				Decls: []js_ast.Decl{{Binding: js_ast.Binding{Loc: nameLoc, Data: &js_ast.BIdentifier{Ref: nameRef}}}},
+				Decls: decls,
 			}})
 		}
 	}
@@ -1165,14 +1440,162 @@ func (p *parser) generateClosureForTypeScriptNamespaceOrEnum(
 		p.recordUsage(nameRef)
 	}
 
+	// Try to use an arrow function if possible for compactness
+	var targetExpr js_ast.Expr
+	args := []js_ast.Arg{{Binding: js_ast.Binding{Loc: nameLoc, Data: &js_ast.BIdentifier{Ref: argRef}}}}
+	if p.options.unsupportedJSFeatures.Has(compat.Arrow) {
+		targetExpr = js_ast.Expr{Loc: stmtLoc, Data: &js_ast.EFunction{Fn: js_ast.Fn{
+			Args: args,
+			Body: js_ast.FnBody{Loc: stmtLoc, Stmts: stmtsInsideClosure},
+		}}}
+	} else {
+		targetExpr = js_ast.Expr{Loc: stmtLoc, Data: &js_ast.EArrow{
+			Args: args,
+			Body: js_ast.FnBody{Loc: stmtLoc, Stmts: stmtsInsideClosure},
+		}}
+	}
+
 	// Call the closure with the name object
 	stmts = append(stmts, js_ast.Stmt{Loc: stmtLoc, Data: &js_ast.SExpr{Value: js_ast.Expr{Loc: stmtLoc, Data: &js_ast.ECall{
-		Target: js_ast.Expr{Loc: stmtLoc, Data: &js_ast.EFunction{Fn: js_ast.Fn{
-			Args: []js_ast.Arg{{Binding: js_ast.Binding{Loc: nameLoc, Data: &js_ast.BIdentifier{Ref: argRef}}}},
-			Body: js_ast.FnBody{Loc: stmtLoc, Stmts: stmtsInsideClosure},
-		}}},
-		Args: []js_ast.Expr{argExpr},
+		Target: targetExpr,
+		Args:   []js_ast.Expr{argExpr},
 	}}}})
 
 	return stmts
 }
+
+func (p *parser) generateClosureForTypeScriptEnum(
+	stmts []js_ast.Stmt, stmtLoc logger.Loc, isExport bool, nameLoc logger.Loc,
+	nameRef js_ast.Ref, argRef js_ast.Ref, exprsInsideClosure []js_ast.Expr,
+	allValuesArePure bool,
+) []js_ast.Stmt {
+	// Bail back to the namespace code for enums that aren't at the top level.
+	// Doing this for nested enums is problematic for two reasons. First of all
+	// enums inside of namespaces must be property accesses off the namespace
+	// object instead of variable declarations. Also we'd need to use "let"
+	// instead of "var" which doesn't allow sibling declarations to be merged.
+	if p.currentScope != p.moduleScope {
+		stmtsInsideClosure := []js_ast.Stmt{}
+		if len(exprsInsideClosure) > 0 {
+			if p.options.mangleSyntax {
+				// "a; b; c;" => "a, b, c;"
+				joined := js_ast.JoinAllWithComma(exprsInsideClosure)
+				stmtsInsideClosure = append(stmtsInsideClosure, js_ast.Stmt{Loc: joined.Loc, Data: &js_ast.SExpr{Value: joined}})
+			} else {
+				for _, expr := range exprsInsideClosure {
+					stmtsInsideClosure = append(stmtsInsideClosure, js_ast.Stmt{Loc: expr.Loc, Data: &js_ast.SExpr{Value: expr}})
+				}
+			}
+		}
+		return p.generateClosureForTypeScriptNamespaceOrEnum(
+			stmts, stmtLoc, isExport, nameLoc, nameRef, argRef, stmtsInsideClosure)
+	}
+
+	// This uses an output format for enums that's different but equivalent to
+	// what TypeScript uses. Here is TypeScript's output:
+	//
+	//   var x;
+	//   (function (x) {
+	//     x[x["y"] = 1] = "y";
+	//   })(x || (x = {}));
+	//
+	// And here's our output:
+	//
+	//   var x = /* @__PURE__ */ ((x) => {
+	//     x[x["y"] = 1] = "y";
+	//     return x;
+	//   })(x || {});
+	//
+	// One benefit is that the minified output is smaller:
+	//
+	//   // Old output minified
+	//   var x;(function(n){n[n.y=1]="y"})(x||(x={}));
+	//
+	//   // New output minified
+	//   var x=(r=>(r[r.y=1]="y",r))(x||{});
+	//
+	// Another benefit is that the @__PURE__ annotation means it automatically
+	// works with tree-shaking, even with more advanced features such as sibling
+	// enum declarations and enum/namespace merges. Ideally all uses of the enum
+	// are just direct references to enum members (and are therefore inlined as
+	// long as the enum value is a constant) and the enum definition itself is
+	// unused and can be removed as dead code.
+
+	// Follow the link chain in case symbols were merged
+	symbol := p.symbols[nameRef.InnerIndex]
+	for symbol.Link != js_ast.InvalidRef {
+		nameRef = symbol.Link
+		symbol = p.symbols[nameRef.InnerIndex]
+	}
+
+	// Generate the body of the closure, including a return statement at the end
+	stmtsInsideClosure := []js_ast.Stmt{}
+	if len(exprsInsideClosure) > 0 {
+		argExpr := js_ast.Expr{Loc: nameLoc, Data: &js_ast.EIdentifier{Ref: argRef}}
+		if p.options.mangleSyntax {
+			// "a; b; return c;" => "return a, b, c;"
+			joined := js_ast.JoinAllWithComma(exprsInsideClosure)
+			joined = js_ast.JoinWithComma(joined, argExpr)
+			stmtsInsideClosure = append(stmtsInsideClosure, js_ast.Stmt{Loc: joined.Loc, Data: &js_ast.SReturn{ValueOrNil: joined}})
+		} else {
+			for _, expr := range exprsInsideClosure {
+				stmtsInsideClosure = append(stmtsInsideClosure, js_ast.Stmt{Loc: expr.Loc, Data: &js_ast.SExpr{Value: expr}})
+			}
+			stmtsInsideClosure = append(stmtsInsideClosure, js_ast.Stmt{Loc: argExpr.Loc, Data: &js_ast.SReturn{ValueOrNil: argExpr}})
+		}
+	}
+
+	// Try to use an arrow function if possible for compactness
+	var targetExpr js_ast.Expr
+	args := []js_ast.Arg{{Binding: js_ast.Binding{Loc: nameLoc, Data: &js_ast.BIdentifier{Ref: argRef}}}}
+	if p.options.unsupportedJSFeatures.Has(compat.Arrow) {
+		targetExpr = js_ast.Expr{Loc: stmtLoc, Data: &js_ast.EFunction{Fn: js_ast.Fn{
+			Args: args,
+			Body: js_ast.FnBody{Loc: stmtLoc, Stmts: stmtsInsideClosure},
+		}}}
+	} else {
+		targetExpr = js_ast.Expr{Loc: stmtLoc, Data: &js_ast.EArrow{
+			Args:       args,
+			Body:       js_ast.FnBody{Loc: stmtLoc, Stmts: stmtsInsideClosure},
+			PreferExpr: p.options.mangleSyntax,
+		}}
+	}
+
+	// Call the closure with the name object and store it to the variable
+	decls := []js_ast.Decl{{
+		Binding: js_ast.Binding{Loc: nameLoc, Data: &js_ast.BIdentifier{Ref: nameRef}},
+		ValueOrNil: js_ast.Expr{Loc: stmtLoc, Data: &js_ast.ECall{
+			Target: targetExpr,
+			Args: []js_ast.Expr{{Loc: nameLoc, Data: &js_ast.EBinary{
+				Op:    js_ast.BinOpLogicalOr,
+				Left:  js_ast.Expr{Loc: nameLoc, Data: &js_ast.EIdentifier{Ref: nameRef}},
+				Right: js_ast.Expr{Loc: nameLoc, Data: &js_ast.EObject{}},
+			}}},
+			CanBeUnwrappedIfUnused: allValuesArePure,
+		}},
+	}}
+	p.recordUsage(nameRef)
+
+	// Use a "var" statement since this is a top-level enum, but only use "export" once
+	stmts = append(stmts, js_ast.Stmt{Loc: stmtLoc, Data: &js_ast.SLocal{
+		Kind:     js_ast.LocalVar,
+		Decls:    decls,
+		IsExport: isExport && !p.emittedNamespaceVars[nameRef],
+	}})
+	p.emittedNamespaceVars[nameRef] = true
+
+	return stmts
+}
+
+func (p *parser) wrapInlinedEnum(value js_ast.Expr, comment string) js_ast.Expr {
+	if p.shouldFoldNumericConstants || p.options.mangleSyntax || strings.Contains(comment, "*/") {
+		// Don't wrap with a comment
+		return value
+	}
+
+	// Wrap with a comment
+	return js_ast.Expr{Loc: value.Loc, Data: &js_ast.EInlinedEnum{
+		Value:   value,
+		Comment: comment,
+	}}
+}
